open Conway
open R313
open Wire
open Spirals
open Starwars
open Unix
open Sys

let cas = ["Conway's Game of Life"; "Cyclic rule 313"; "Wire World"; "Spirals"; "Star Wars"]
let set = ref 1

let endrun mainwin runwin =
  runwin#destroy ();
  mainwin#show ()

let configure buffer ev =
  let width = GdkEvent.Configure.width ev in
  let height = GdkEvent.Configure.height ev in
  let pixmap = GDraw.pixmap ~width ~height () in
  pixmap#set_foreground `WHITE;
  pixmap#rectangle ~x:0 ~y:0 ~width ~height ~filled:true ();
  buffer := pixmap;
  true

let expose (draw:GMisc.drawing_area) (buffer:GDraw.pixmap ref) ev =
  let area = GdkEvent.Expose.area ev in
  let x = Gdk.Rectangle.x area in
  let y = Gdk.Rectangle.y area in
  let width = Gdk.Rectangle.width area in
  let height = Gdk.Rectangle.width area in
  let drawing =
    draw#misc#realize ();
    new GDraw.drawable (draw#misc#window)
  in drawing#put_pixmap ~x ~y ~xsrc:x ~ysrc:y ~width ~height !buffer#pixmap;
  false

let set_cell (draw:GMisc.drawing_area) (buffer:GDraw.pixmap ref) x y z state colors =
  let x = x*z and y = y*z in
  let update_rect = Gdk.Rectangle.create ~x ~y ~width:z ~height:z in
  !buffer#set_foreground colors.(state);
  !buffer#rectangle ~x ~y ~width:z ~height:z ~filled:true ();
  draw#misc#draw (Some update_rect)

let refresh draw (buffer:GDraw.pixmap ref) z cao colors =
  let w = cao#get_width and h = cao#get_height in
  for i=0 to (w-1) do
    for j=0 to (h-1) do
      let x = i*z and y = j*z in
      !buffer#set_foreground colors.((cao#get_cell i j)#get_state);
      !buffer#rectangle ~x ~y ~width:z ~height:z ~filled:true ()
    done
  done ;
  let drawing = draw#misc#realize (); new GDraw.drawable (draw#misc#window)
  in drawing#put_pixmap ~x:0 ~y:0 ~xsrc:0 ~ysrc:0 ~width:(z*w) ~height:(z*h) !buffer#pixmap

let button_pressed draw buffer z cao colors ev =
  if GdkEvent.Button.button ev = 1 then (
    let x = (int_of_float (GdkEvent.Button.x ev))/z in
    let y = (int_of_float (GdkEvent.Button.y ev))/z in
    set_cell draw buffer x y z !set colors;
    cao#set_cell x y (new Ca.cell !set)
  );
  true

let runca mainwin ca w h z =
  let cao = match ca with
    | 0 -> new conway w h
    | 1 -> new r313 w h
    | 2 -> new wire w h
    | 3 -> new spirals w h
    | 4 -> new starwars w h
    | _ -> failwith "Automaton not found !" in
  let rec avs = function 0->[] | i->(string_of_int (cao#get_states-i))::(avs (i-1)) in
  let colors = [| `WHITE; `BLACK; `NAME "blue" ; `NAME "red" ; `NAME "yellow" ; `NAME "green" |] in

  let runwin = GWindow.window ~title:"Cellular Automata Simulator"
    ~resizable:false ~border_width:6 ~position:`CENTER () ~icon:(GdkPixbuf.from_file "ca.xpm") in
  let mainbox = GPack.vbox ~packing:runwin#add () in
  let bbox = GPack.hbox ~homogeneous:false ~spacing:5
    ~packing:mainbox#pack () in
  let qbutton = GButton.button ~label:"Return"
    ~packing:bbox#pack () in

  let slabel = GMisc.label ~text:"Set cell to"
    ~packing:bbox#pack () in
  let slist = GEdit.combo_box_text ~strings:(avs cao#get_states)
    ~packing:bbox#pack () in (fst slist)#set_active 1;
  (fst slist)#connect#changed ~callback:(fun ()->set:=(fst slist)#active);

  let rbutton = GButton.button ~label:"Random state"
    ~packing:bbox#pack () in
  let cbutton = GButton.button ~label:"Reset state"
    ~packing:bbox#pack () in

  let _ = GMisc.separator `HORIZONTAL
    ~packing:(mainbox#pack ~expand:false ~fill:true ~padding:5) () in
  let cbox = GPack.hbox ~homogeneous:true ~spacing:3
    ~packing:mainbox#pack () in
  let clabel = GMisc.label ~text:"Colors :"
    ~packing:cbox#pack () in

  let _ = GMisc.separator `HORIZONTAL
    ~packing:(mainbox#pack ~expand:false ~fill:true ~padding:5) () in
  let sbox = GPack.hbox ~homogeneous:true ~spacing:3
    ~packing:mainbox#pack () in

  let sbutton = GButton.button ~label:"Step"
    ~packing:(sbox#pack ~expand:true ~fill:true) () in
  let _ = GMisc.separator `VERTICAL
    ~packing:(sbox#pack ~padding:5) () in
  let wlabel = GMisc.label ~text:"Animate :"
    ~packing:sbox#pack () in
  let aadj = GData.adjustment ~value:0.1 ~lower:0.05 ~upper:1. ~step_incr:0.01 ~page_incr:0.01 ~page_size:0. () in
  let asel = GRange.scale `HORIZONTAL ~adjustment:aadj ~digits:2 ~draw_value:true ~value_pos:`LEFT
    ~packing:(sbox#pack ~fill:true ~expand:true) () in
  let abutton = GButton.button ~label:"Go"
    ~packing:(sbox#pack ~expand:false ~fill:false) () in

  let _ = GMisc.separator `HORIZONTAL
    ~packing:(mainbox#pack ~expand:false ~fill:true ~padding:5) () in

  let table = GPack.table ~rows:2 ~columns:2 ~packing:mainbox#add () in
  let draw = GMisc.drawing_area ~width:(w*z) ~height:(h*z)
    ~packing:(table#attach ~left:1 ~top:1) () in
  draw#event#add [`POINTER_MOTION; `POINTER_MOTION_HINT; `EXPOSURE;
    `LEAVE_NOTIFY; `BUTTON_PRESS ];

  let hruler = GRange.ruler `HORIZONTAL ~metric:`PIXELS
    ~lower:0.0 ~upper:(float w) ~position:0.0 ~max_size:(float w)
    ~packing:(table#attach ~left:1 ~top:0) () in
  draw#event#connect#motion_notify
    ~callback:(fun ev -> hruler#event#send (ev :> GdkEvent.any));
  let vruler = GRange.ruler `VERTICAL ~metric:`PIXELS
    ~lower:0.0 ~upper:(float h) ~position:0.0 ~max_size:(float h)
    ~packing:(table#attach ~left:0 ~top:1) () in
  draw#event#connect#motion_notify
    ~callback:(fun ev -> vruler#event#send (ev :> GdkEvent.any));
  let buffer = ref (GDraw.pixmap ~width:(w*z) ~height:(h*z) ()) in

  draw#event#connect#expose ~callback:(expose draw buffer);
  draw#event#connect#configure ~callback:(configure buffer);
  draw#event#connect#button_press ~callback:(button_pressed draw buffer z cao colors);
  sbutton#connect#clicked ~callback:(fun ()->cao#step;refresh draw buffer z cao colors);
  rbutton#connect#clicked ~callback:(fun ()->cao#random_configuration;refresh draw buffer z cao colors);
  cbutton#connect#clicked ~callback:(fun ()->cao#reset;refresh draw buffer z cao colors);

  let anim = ref false in
  let toggle_anim () = if !anim then (anim := false; abutton#set_label "Go";
      setitimer ITIMER_REAL {it_interval=0.0; it_value=0.0};
      set_signal sigalrm Signal_default;)
    else
      let t = {it_interval = aadj#value; it_value = aadj#value} in
      (setitimer ITIMER_REAL t; abutton#set_label "Stop"; anim := true;
      set_signal sigalrm (Signal_handle (fun _->cao#step;refresh draw buffer z cao colors;)))
  in abutton#connect#clicked ~callback:(toggle_anim);

  for i=0 to cao#get_states-1 do
    let button = GButton.color_button ~color:(GDraw.color colors.(i)) ~packing:cbox#pack () in
    button#connect#color_set ~callback:(fun ()->let nc = button#color in
      colors.(i) <- `RGB (Gdk.Color.red nc, Gdk.Color.green nc, Gdk.Color.blue nc))
  done;

  runwin#event#connect#delete ~callback:(fun _->if (!anim) then toggle_anim (); endrun mainwin runwin;true);
  qbutton#connect#clicked ~callback:(fun ()->if(!anim) then toggle_anim (); endrun mainwin runwin);
  mainwin#misc#hide ();
  runwin#show ()

let main () =
  let window = GWindow.window ~title:"Cellular Automata Simulator" ~width:500
    ~resizable:false ~border_width:6 ~position:`CENTER () ~icon:(GdkPixbuf.from_file "ca.xpm") in
  let mainbox = GPack.vbox ~packing:window#add () in

  let selectbox = GPack.hbox ~homogeneous:false ~spacing:10
    ~packing:(mainbox#pack ~padding:2) () in
  let calabel = GMisc.label ~text:"Automaton :"
    ~packing:selectbox#pack () in
  let calist = GEdit.combo_box_text ~strings:cas
    ~packing:(selectbox#pack ~fill:true ~expand:true) () in

  let zlabel = GMisc.label ~text:"Pixels per cell :"
    ~packing:selectbox#pack () in
  let zlist = GEdit.combo_box_text ~strings:["1";"2";"3";"4";"5";"6"]
    ~packing:selectbox#pack () in

  let _ = GMisc.separator `HORIZONTAL
    ~packing:(mainbox#pack ~expand:false ~fill:true ~padding:5) () in

  let dimbox = GPack.hbox ~homogeneous:false ~spacing:10
    ~packing:(mainbox#pack ~padding:2) () in
  let wlabel = GMisc.label ~text:"Width :"
    ~packing:dimbox#pack () in
  let wadj = GData.adjustment ~value:100. ~lower:50. ~upper:200. ~step_incr:1. ~page_incr:1. ~page_size:0. () in
  let wsel = GRange.scale `HORIZONTAL ~adjustment:wadj ~digits:0 ~draw_value:true ~value_pos:`LEFT
    ~packing:(dimbox#pack ~fill:true ~expand:true) () in

  let hlabel = GMisc.label ~text:"Height :"
    ~packing:dimbox#pack () in
  let hadj = GData.adjustment ~value:100. ~lower:50. ~upper:200. ~step_incr:1. ~page_incr:1. ~page_size:0. () in
  let hsel = GRange.scale `HORIZONTAL ~adjustment:hadj ~digits:0 ~draw_value:true ~value_pos:`LEFT
    ~packing:(dimbox#pack ~fill:true ~expand:true) () in

  let _ = GMisc.separator `HORIZONTAL
    ~packing:(mainbox#pack ~expand:false ~fill:true ~padding:5) () in

  let bbox = GPack.hbox ~homogeneous:true ~spacing:3
    ~packing:mainbox#pack () in
  let qbutton = GButton.button ~label:"Quit"
    ~packing:(bbox#pack ~expand:true ~fill:true) () in
  let rbutton = GButton.button ~label:"Run"
    ~packing:(bbox#pack ~expand:true ~fill:true) () in

  window#connect#destroy ~callback:GMain.Main.quit;
  qbutton#connect#clicked ~callback:GMain.Main.quit;
  rbutton#connect#clicked ~callback:(fun ()->runca window (fst calist)#active
    (int_of_float wadj#value) (int_of_float hadj#value) (1+(fst zlist)#active));
  (fst calist)#set_active 0; (fst zlist)#set_active 3; window#show ();
  GMain.Main.main ()

let _ = Printexc.print main ()
